//Designed by Christopher Mason for Dr. David McAllister
//Version date 8-16-02
//Compiled only for windows but code should be compatable with other OSs
//Red Book refers to the OpenGL Programming Guide 3rd Edition by Woo, Neider, Davis, and Shreiner
//White Pages refers to the information at http://www.stereographics.com/developer_program/pcsdk.htm

#include<GL/GLUT.H>
#include<GL/GL.H>
#include<GL/glui.h>
#include<stdlib.h>

typedef struct {
	double x, y, z;
} vector;

void init(void);
//Initilizes the scene's values
void display(void);
//Displays the scene
void reshape(int w, int h);
//Resizes the scene to fit the window when resized
void keyboard(unsigned char key, int x, int y);
//Takes in keys that are pressed
void myGlutIdle(void);
//The idle function
void drawScene(bool picking);
//Draws the scene, call in red and blue-green channel and picking
//bool picking - set to true during picking to give objects names so they picking routines will work properly
void updateMap(void);
//Updates the bezier surface, call when any control point values are changed
void mouse(int button, int state, int x, int y);
//Function that processes mouse clicks, information on the selection routine used can be found in the red book
//on p540-550
void motion(int x, int y);
//Function that processes mouse movement when clicked.

int mainWindow;						//Value of the main graphics window

float view_rotate[16] = { 1,0,0,0, 0,1,0,0, 0,0,1,0, 0,0,0,1 };		//Identity matrix for the rotation object
float view_translate = 0.0;			//Base value of 0 for the zoom amount

GLfloat ctrlpoints[4][4][3] = {
	{{-300.0, -300.0, 0.0}, {-100.0, -300.0, 0.0},
	{100.0, -300.0, 0.0}, {300.0, -300.0, 0.0}},
	{{-300.0, -100.0, 0.0}, {-100.0, -100.0, 0.0},
	{100.0, -100.0, 0.0}, {300.0, -100.0, 0.0}},
	{{-300.0, 100.0, 0.0}, {-100.0, 100.0, 0.0},
	{100.0, 100.0, 0.0}, {300.0, 100.00, 0.0}},
	{{-300.0, 300.0, 0.0}, {-100.0, 300.0, 0.0},
	{100.0, 300.0, 0.0}, {300.0, 300.0, 0.0}}
};									//Starting location of the control points
int pickedPointX;					//Number on the x axis of the selected point -1 if a point is not selected
int pickedPointY;					//Same except on y axis
int pressed;						//Identifies what button was pressed
int mouseX;							//Stores last X loaction of the mouse
int mouseY;							//Stores last Y location of the mouse
float offsetAdj = 1.0;				//Default value for the scale of the offset of the windows (see white pages section 7)
float balanceAdj = 1.0;				//Default value for the balance (see white pages section 7)
float ratio = 1.0 / 700.0;			//Ratio of width to distance.

int main(int argc, char *argv[]) {
	
	//Setup the window and it's commands (basic stuff)
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowPosition(200, 0);
	glutInitWindowSize(500, 500);
	mainWindow = glutCreateWindow("Bezier Surface.");
	init();
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutMouseFunc(mouse);
	glutMotionFunc(motion);

	//Setup Menu with rotation control and zoom control
	GLUI *glui = GLUI_Master.create_glui("Controls", 0, 0, 0);
		GLUI_Rollout *View_rollout = glui->add_rollout("View Controls", true);
			GLUI_Rotation *View_rot = glui->add_rotation_to_panel(View_rollout, "Rotate:", view_rotate);
				View_rot->set_spin(1.0);
			GLUI_Translation *View_trans = glui->add_translation_to_panel(View_rollout, "Zoom:", GLUI_TRANSLATION_Z, &view_translate);
				View_trans->set_speed(10.0);

			//Not needed unless someone has trouble with the defaults, highly unlikely 
//			GLUI_Spinner *Offset_spinner = glui->add_spinner_to_panel(View_rollout, "Offset:", GLUI_SPINNER_FLOAT, &offsetAdj);
//				Offset_spinner->set_float_limits(0.0, 2.0);
//			GLUI_Spinner *Balance_spinner = glui->add_spinner_to_panel(View_rollout, "Balance:", GLUI_SPINNER_FLOAT, &balanceAdj);
//				Balance_spinner->set_float_limits(0.0, 2.0);
				

	//Let glui know what the gfx window is and initilize it's idle function.
	glui->set_main_gfx_window(mainWindow);
	GLUI_Master.set_glutIdleFunc(myGlutIdle);

	//Start the main loop
	glutMainLoop();
	return 0;
}

void myGlutIdle(void) {
	
	if(glutGetWindow() != mainWindow) 
		glutSetWindow(mainWindow);
	glutPostRedisplay();
}

void init(void) {
	//Initilize main values.
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glEnable(GL_NORMALIZE);
	glEnable(GL_DEPTH_TEST);
	glLineWidth(3.0);
	glColor3ub(200, 200, 200);
	//Update the surface
	updateMap();

}

void display(void) {
	
	//Calculate the amount to offset and balance (see white pages section 7)
	float offset = 652.83 * 0.035 * offsetAdj;
	float balance = -offset * balanceAdj * ratio;
	glClearColor(0.0, 0.0, 0.0, 1.0);
	glColorMask(1.0, 1.0, 1.0, 1.0);
	glClear(GL_COLOR_BUFFER_BIT);
	glClearColor(1.0, 1.0, 1.0, 1.0);

	//Blue-Green Channel
	//Draw the scene from the right camera useing the offsets and balance (see white pages section 6)
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-0.4663+balance, 0.4663+balance, -0.4663, 0.4663, 1.0, 30000.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	//Set the color mask to only allow blue and green values to be changed.
	glColorMask(0.0, 1.0, 1.0, 0.0);
	gluLookAt(0.0, 0.0, 700.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	glTranslatef(-offset, 0.0, 0.0);
	drawScene(false);

	//Red Channel
	//Draw the scene from the right camera useing the offsets and balance (see white pages section 6)
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glFrustum(-0.4663-balance, 0.4663-balance, -0.4663, 0.4663, 1.0, 30000.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	//Set the color mask to only allow red values to be changed.
	glColorMask(1.0, 0.0, 0.0, 0.0);
	gluLookAt(0.0, 0.0, 700.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	glTranslatef(offset, 0.0, 0.0);
	drawScene(false);

	//Display the scene
	glutSwapBuffers(); 
}

void drawScene(bool picking) {
	
	int j, i;	//For loops
	//Clear the buffers from the old frame
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	//Push the current matrix to the stack then translate for zoom and rotate the world
	glPushMatrix();
	glTranslatef(0.0, 0.0, view_translate);
	glMultMatrixf(view_rotate);

	//Do not draw grid during picking
	if(picking != true) {
		//Standard code for drawing surfaces (Red Book p504)
		for(j = 0; j <= 3; j++) {
			glBegin(GL_LINE_STRIP);
			for(i = 0; i <= 40; i++) {
				glEvalCoord2f((GLfloat)i/40.0, (GLfloat)j/3.0);	
			}
			glEnd();
			glBegin(GL_LINE_STRIP);
			for(i = 0; i <= 40; i++) {
				glEvalCoord2f((GLfloat)j/3.0, (GLfloat)i/40.0);	
			}
			glEnd();
		}
	}

	//Load names only during picking
	if(picking == true)
		for(int n = 0; n < 16; n++) {
			glLoadName(i);
		}
	
	//Code to draw the control points.
	for(j = 0; j < 4; j++) {
		for(i = 0; i < 4; i++) {
			glPushMatrix();
			glTranslatef(ctrlpoints[i][j][0], ctrlpoints[i][j][1], ctrlpoints[i][j][2]);

			//Only give name when picking
			if(picking == true)
				glPushName((j*4)+i);
			glutSolidCube(10.0);

			//Only needed during picking
			if(picking == true)
				glPopName();
			glPopMatrix();
		}
	}
	glPopMatrix();
}

void mouse(int button, int state, int x, int y) {
	
	if(state == GLUT_DOWN) {
			pressed = button;	//Stores the button that was pressed
			mouseX = x;			//Stores the mouse's x position for motion
			mouseY = y;			//Stores the mouse's y position for motion
			GLint viewport[4];	//Array for viewport information
			GLuint selectBuf[64];	//Buffer for slection information
			GLint numHits;			//Stores the number of hits for each selection
			GLfloat min = 2.0;		//Z value for the closest point initilized to -2.0 because all values are
									//by default between 0.0 and 1.0 so this number will be replaced by the first one
			GLint minName = -1;		//Name of the point that has the minimum z value and therefore closest to the viewer

			//Get viewport information and set the selection buffer
			glGetIntegerv(GL_VIEWPORT, viewport);
			glSelectBuffer(64, selectBuf);

			//Go into rendermode and initilize the names
			glRenderMode(GL_SELECT);
			glInitNames();
			glPushName(0);
			
			//Set up the scene to not draw in stereo
			glClearColor(1.0, 1.0, 1.0, 0.0);
			glColorMask(1.0, 1.0, 1.0, 0.0);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glMatrixMode(GL_PROJECTION);
			glPushMatrix();
			glLoadIdentity();
			gluPickMatrix(x, viewport[3]-y, 1.0, 1.0, viewport);
			glFrustum(-0.4663, 0.4663, -0.4663, 0.4663, 1.0, 30000.0);
			glMatrixMode(GL_MODELVIEW);
			glLoadIdentity();
			gluLookAt(0.0, 0.0, 700.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
			drawScene(true);
			glPopMatrix();
			glutSwapBuffers();
			
			//Store the number of hits and determine the one closest to the viewer
			numHits = glRenderMode(GL_RENDER);
			if(numHits != 0) {
				for(int i = 0; i < numHits; i++) {
					if((float)selectBuf[i*5+1]/0x7fffffff < min) {
						min = (float)selectBuf[i*5+1]/0x7fffffff;
						minName = selectBuf[i*5+4];
					}
				}
				//Store the closest point
				pickedPointX = minName % 4;
				pickedPointY = minName / 4;
			}
			else {
				//-1 are not valid so no point was selected here.
				pickedPointX = -1;
				pickedPointY = -1;
			}
	}
}

void motion(int x, int y) {

	
	int movementX = x - mouseX;			//Amount moved in X
	int movementY = -1 * (y - mouseY);	//Amount moved in Y
	float trans[3];						//Amounts to translate the control points by

	//If no control point is selected or the mouse has not moved get out of function
	if(movementX == 0 && movementY == 0 || pickedPointX == -1)
		return;

	//Left button presses are movement in X and Y
	if(pressed == GLUT_LEFT_BUTTON) {
		trans[0] = view_rotate[0]*movementX + view_rotate[4]*movementY;
		trans[1] = view_rotate[4]*movementX + view_rotate[5]*movementY;
		trans[2] = view_rotate[8]*movementX + view_rotate[9]*movementY;

	}
	//Right button presses are movment in X and Z
	else if(pressed == GLUT_RIGHT_BUTTON) {
		trans[0] = view_rotate[0]*movementX + view_rotate[2]*movementY;
		trans[1] = view_rotate[4]*movementX + view_rotate[6]*movementY;
		trans[2] = view_rotate[8]*movementX + view_rotate[10]*movementY;
	}
	//Other buttons are ignored
	else
		return;

	//Translate the picked control point
	ctrlpoints[pickedPointX][pickedPointY][0] += trans[0];
	ctrlpoints[pickedPointX][pickedPointY][1] += trans[1];
	ctrlpoints[pickedPointX][pickedPointY][2] += trans[2];
	
	//Store the current mouse position to calculate movement amount next time through.
	mouseX = x;
	mouseY = y;

	//Update the surface
	updateMap();
}

void reshape(int w, int h) {
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
}

void updateMap(void) {
	
	//Update the surface map(Red Book p504)
    glMap2f(GL_MAP2_VERTEX_3, 0, 1, 3, 4, 0, 1, 12, 4, &ctrlpoints[0][0][0]);
	glEnable(GL_MAP2_VERTEX_3);

}

